from jax import jit
import jax.numpy as jnp
import torch
import numpy as np
from typing import Tuple
from functools import partial
from s2wav import samples


def filters_axisym(
    L: int, J_min: int = 0, lam: float = 2.0
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Computes wavelet kernels :math:`\Psi^j_{\ell m}` and scaling kernel
        :math:`\Phi_{\ell m}` in harmonic space.

    Specifically, these kernels are derived in `[1] <https://arxiv.org/pdf/1211.1680.pdf>`_,
    where the wavelet kernels are defined (15) for scale :math:`j` to be

    .. math::

        \Psi^j_{\ell m} \equiv \sqrt{\frac{2\ell+1}{4\pi}} \kappa_{\lambda}(\frac{\ell}{\lambda^j})\delta_{m0},

    where :math:`\kappa_{\lambda} = \sqrt{k_{\lambda}(t/\lambda) - k_{\lambda}(t)}` for :math:`k_{\lambda}`
    given in :func:`~k_lam`. Similarly, the scaling kernel is defined (16) as

    .. math::

        \Phi_{\ell m} \equiv \sqrt{\frac{2\ell+1}{4\pi}} \nu_{\lambda} (\frac{\ell}{\lambda^{J_0}})\delta_{m0},

    where :math:`\nu_{\lambda} = \sqrt{k_{\lambda}(t)}` for :math:`k_{\lambda}` given in :func:`~k_lam`.
    Notice that :math:`\delta_{m0}` enforces that these kernels are axisymmetric, i.e. coefficients
    for :math:`m \not = \ell` are zero. In this implementation the normalisation constant has been
    omitted as it is nulled in subsequent functions.

    Args:
        L (int): Harmonic band-limit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

    Raises:
        ValueError: J_min is negative or greater than J.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Unnormalised wavelet kernels :math:`\Psi^j_{\ell m}`
        with shape :math:`[(J+1)L]`, and scaling kernel :math:`\Phi_{\el m}` with shape
        :math:`[L]` in harmonic space.

    Note:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the sphere", A&A, vol. 558, p. A128, 2013.
    """
    J = samples.j_max(L, lam)

    if J_min >= J or J_min < 0:
        raise ValueError(
            "J_min must be non-negative and less than J= "
            + str(J)
            + " for given L and lam."
        )

    previoustemp = 0.0
    k = k_lam(L, lam)
    psi = np.zeros((J + 1, L), np.float64)
    phi = np.zeros(L, np.float64)
    for l in range(L):
        phi[l] = np.sqrt(k[J_min, l])

    for j in range(J_min, J + 1):
        for l in range(L):
            diff = k[j + 1, l] - k[j, l]
            if diff < 0:
                psi[j, l] = previoustemp
            else:
                temp = np.sqrt(diff)
                psi[j, l] = temp
            previoustemp = temp

    return psi, phi


def filters_directional(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    using_torch: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Generates the harmonic coefficients for the directional tiling wavelets.

    This implementation is based on equation 36 in the wavelet computation paper
    `[1] <https://arxiv.org/pdf/1509.06749.pdf>`_.

    Args:
        L (int): Harmonic band-limit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) to perform the transform. Defaults to 0.

        spin0 (int, optional): Spin number the wavelet was lowered from. Defaults to 0.

        using_torch (bool, optional): Desired frontend functionality. Defaults to False.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Tuple of wavelet and scaling kernels
            (:math:`\Psi^j_{\ell n}`, :math:`\Phi_{\ell m}`)

    Notes:
        [1] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint arXiv:1509.06749 (2015).
    """
    J = samples.j_max(L, lam)
    el_min = max(abs(spin), abs(spin0))

    phi = np.zeros(L, dtype=np.float64)
    psi = np.zeros((J + 1, L, 2 * L - 1), dtype=np.complex128)

    kappa, kappa0 = filters_axisym(L, J_min, lam)
    s_elm = tiling_direction(L, N)

    for el in range(el_min, L):
        if kappa0[el] != 0:
            phi[el] = np.sqrt((2 * el + 1) / (4.0 * np.pi)) * kappa0[el]
            if spin0 != 0:
                phi[el] *= _spin_normalization(el, spin0) * (-1) ** spin0

    for j in range(J_min, J + 1):
        for el in range(el_min, L):
            if kappa[j, el] != 0:
                for m in range(-el, el + 1):
                    if s_elm[el, L - 1 + m] != 0:
                        psi[j, el, L - 1 + m] = (
                            np.sqrt((2 * el + 1) / (8.0 * np.pi * np.pi))
                            * kappa[j, el]
                            * s_elm[el, L - 1 + m]
                        )
                        if spin0 != 0:
                            psi[j, el, L - 1 + m] *= (
                                _spin_normalization(el, spin0) * (-1) ** spin0
                            )
    if using_torch:
        psi = torch.from_numpy(psi)
        phi = torch.from_numpy(phi)

    return psi, phi


def filters_axisym_vectorised(
    L: int, J_min: int = 0, lam: float = 2.0
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Vectorised version of :func:`~filters_axisym`.

    Args:
        L (int): Harmonic band-limit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales. Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

    Raises:
        ValueError: J_min is negative or greater than J.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Unnormalised wavelet kernels :math:`\Psi^j_{\ell m}`
        with shape :math:`[(J+1)L], and scaling kernel :math:`\Phi_{\ell m}` with shape
        :math:`[L]` in harmonic space.
    """
    J = samples.j_max(L, lam)

    if J_min >= J or J_min < 0:
        raise ValueError(
            "J_min must be non-negative and less than J= "
            + str(J)
            + " for given L and lam."
        )

    k = k_lam(L, lam)
    diff = (np.roll(k, -1, axis=0) - k)[:-1]
    diff[diff < 0] = 0
    return np.sqrt(diff), np.sqrt(k[J_min])


def filters_directional_vectorised(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    using_torch: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Vectorised version of :func:`~filters_directional`.

    Args:
        L (int): Harmonic band-limit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) to perform the transform. Defaults to 0.

        spin0 (int, optional): Spin number the wavelet was lowered from. Defaults to 0.

        using_torch (bool, optional): Desired frontend functionality. Defaults to False.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Tuple of wavelet and scaling kernels
            (:math:`\Psi^j_{\ell n}`, :math:`\Phi_{\ell m}`).
    """
    el_min = max(abs(spin), abs(spin0))

    spin_norms = (
        (-1) ** spin0 * _spin_normalization_vectorised(np.arange(L), spin0)
        if spin0 != 0
        else 1
    )

    kappa, kappa0 = filters_axisym_vectorised(L, J_min, lam)
    s_elm = tiling_direction(L, N)

    kappa0 *= np.sqrt((2 * np.arange(L) + 1) / (4.0 * np.pi))
    kappa0 = kappa0 * spin_norms if spin0 != 0 else kappa0

    kappa *= np.sqrt((2 * np.arange(L) + 1) / 8.0) / np.pi
    kappa = np.einsum("ij,jk->ijk", kappa, s_elm)
    kappa = np.einsum("ijk,j->ijk", kappa, spin_norms) if spin0 != 0 else kappa

    kappa0[:el_min] = 0
    kappa[:, :el_min, :] = 0

    if using_torch:
        kappa0 = torch.from_numpy(kappa0)
        kappa = torch.from_numpy(kappa)

    return kappa, kappa0


@partial(jit, static_argnums=(0, 1, 2))
def filters_axisym_jax(
    L: int, J_min: int = 0, lam: float = 2.0
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    r"""JAX version of :func:`~filters_axisym_vectorised`.

    Args:
        L (int): Harmonic band-limit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales. Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

    Raises:
        ValueError: J_min is negative or greater than J.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Unnormalised wavelet kernels :math:`\Psi^j_{\ell m}`
        with shape :math:`[(J+1)L], and scaling kernel :math:`\Phi_{\ell m}` with shape
        :math:`[L]` in harmonic space.
    """
    J = samples.j_max(L, lam)

    if J_min >= J or J_min < 0:
        raise ValueError(
            "J_min must be non-negative and less than J= "
            + str(J)
            + " for given L and lam."
        )

    k = k_lam_jax(L, lam)
    diff = (jnp.roll(k, -1, axis=0) - k)[:-1]
    diff = jnp.where(diff < 0, jnp.zeros((J + 1, L)), diff)
    return jnp.sqrt(diff), jnp.sqrt(k[J_min])


@partial(jit, static_argnums=(0, 1, 2, 3, 4, 5))
def filters_directional_jax(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    r"""JAX version of :func:`~filters_directional`.

    Args:
        L (int): Harmonic band-limit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) to perform the transform. Defaults to 0.

        spin0 (int, optional): Spin number the wavelet was lowered from. Defaults to 0.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Tuple of wavelet and scaling kernels
            (:math:`\Psi^j_{\ell n}`, :math:`\Phi_{\ell m}`).
    """
    el_min = max(abs(spin), abs(spin0))

    spin_norms = (
        (-1) ** spin0 * _spin_normalization_jax(np.arange(L), spin0)
        if spin0 != 0
        else 1
    )

    kappa, kappa0 = filters_axisym_jax(L, J_min, lam)
    s_elm = tiling_direction_jax(L, N)

    kappa0 *= jnp.sqrt((2 * jnp.arange(L) + 1) / (4.0 * jnp.pi))
    kappa0 = kappa0 * spin_norms if spin0 != 0 else kappa0

    kappa *= jnp.sqrt((2 * jnp.arange(L) + 1) / 8.0) / np.pi
    kappa = jnp.einsum("ij,jk->ijk", kappa, s_elm, optimize=True)
    kappa = (
        jnp.einsum("ijk,j->ijk", kappa, spin_norms, optimize=True)
        if spin0 != 0
        else kappa
    )

    kappa0 = kappa0.at[:el_min].set(0)
    kappa = kappa.at[:, :el_min, :].set(0)

    return kappa, kappa0


def tiling_integrand(t: float, lam: float = 2.0) -> float:
    r"""Tiling integrand for scale-discretised wavelets `[1] <https://arxiv.org/pdf/1211.1680.pdf>`_.

    Intermediate step used to compute the wavelet and scaling function generating
    functions. One of the basic mathematical functions needed to carry out the tiling of
    the harmonic space.

    Args:
        t (float): Real argument over which we integrate.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales.Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

    Returns:
        float: Value of tiling integrand for given :math:`t` and scaling factor.

    Note:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on
            the sphere", A&A, vol. 558, p. A128, 2013.
    """
    s_arg = (t - (1.0 / lam)) * (2.0 * lam / (lam - 1.0)) - 1.0

    integrand = np.exp(-2.0 / (1.0 - s_arg**2.0)) / t

    return integrand


def part_scaling_fn(a: float, b: float, n: int, lam: float = 2.0) -> float:
    r"""Computes integral used to calculate smoothly decreasing function :math:`k_{\lambda}`.

    Intermediate step used to compute the wavelet and scaling function generating
    functions. Uses the trapezium method to integrate :func:`~tiling_integrand` in the
    limits from :math:`a \rightarrow b` with scaling parameter :math:`\lambda`. One of
    the basic mathematical functions needed to carry out the tiling of the harmonic
    space.

    Args:
        a (float): Lower limit of the numerical integration.

        b (float): Upper limit of the numerical integration.

        n (int): Number of steps to be performed during integration.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales.Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

    Returns:
        float: Integral of the tiling integrand from :math:`a \rightarrow b`.
    """
    sum = 0.0
    h = (b - a) / n

    if a == b:
        return 0

    for i in range(n):
        if a + i * h not in [1 / lam, 1.0] and a + (i + 1) * h not in [
            1 / lam,
            1.0,
        ]:
            f1 = tiling_integrand(a + i * h, lam)
            f2 = tiling_integrand(a + (i + 1) * h, lam)

            sum += ((f1 + f2) * h) / 2

    return sum


def k_lam(L: int, lam: float = 2.0, quad_iters: int = 300) -> float:
    r"""Compute function :math:`k_{\lambda}` used as a wavelet generating function.

    Specifically, this function is derived in [1] and is given by

    .. math::

        k_{\lambda} \equiv \frac{ \int_t^1 \frac{\text{d}t^{\prime}}{t^{\prime}}
        s_{\lambda}^2(t^{\prime})}{ \int_{\frac{1}{\lambda}}^1
        \frac{\text{d}t^{\prime}}{t^{\prime}} s_{\lambda}^2(t^{\prime})},

    where the integrand is defined to be

    .. math::

        s_{\lambda} \equiv s \Big ( \frac{2\lambda}{\lambda - 1}(t-\frac{1}{\lambda})
        - 1 \Big ),

    for infinitely differentiable Cauchy-Schwartz function :math:`s(t) \in C^{\infty}`.

    Args:
        L (int): Harmonic band-limit.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales. Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

        quad_iters (int, optional): Total number of iterations for quadrature
            integration. Defaults to 300.

    Returns:
        (np.ndarray): Value of :math:`k_{\lambda}` computed for values between
            :math:`\frac{1}{\lambda}` and 1, parametrised by :math:`\ell` as required to
            compute the axisymmetric filters in :func:`~tiling_axisym`.

    Note:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the
            sphere", A&A, vol. 558, p. A128, 2013.
    """

    J = samples.j_max(L, lam)

    normalisation = part_scaling_fn(1.0 / lam, 1.0, quad_iters, lam)
    k = np.zeros((J + 2, L))

    for j in range(J + 2):
        for l in range(L):
            if l < lam ** (j - 1):
                k[j, l] = 1
            elif l > lam**j:
                k[j, l] = 0
            else:
                k[j, l] = (
                    part_scaling_fn(l / lam**j, 1.0, quad_iters, lam) / normalisation
                )

    return k


@partial(jit, static_argnums=(2, 3))  # not sure
def _part_scaling_fn_jax(a: float, b: float, n: int, lam: float = 2.0) -> float:
    r"""JAX version of part_scaling_fn. Computes integral used to calculate smoothly
        decreasing function :math:`k_{\lambda}`.

    Intermediate step used to compute the wavelet and scaling function generating
    functions. Uses the trapezium method to integrate :func:`~tiling_integrand` in the
    limits from :math:`a \rightarrow b` with scaling parameter :math:`\lambda`. One of
    the basic mathematical functions needed to carry out the tiling of the harmonic
    space.

    Args:
        a (float): Lower limit of the numerical integration.

        b (float): Upper limit of the numerical integration.

        n (int): Number of steps to be performed during integration.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales.Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

    Returns:
        float: Integral of the tiling integrand from :math:`a \rightarrow b`.
    """

    h = (b - a) / n

    x = jnp.linspace(a, b, num=n + 1)
    s_arg = (x - (1.0 / lam)) * (2.0 * lam / (lam - 1.0)) - 1.0
    value = jnp.where(
        (x[:-1] == 1.0 / lam) | (x[:-1] == 1.0) | (x[1:] == 1.0 / lam) | (x[1:] == 1.0),
        jnp.zeros(n),
        (jnp.exp(-2.0 / (1.0 - jnp.square(s_arg))) / x)[:-1]
        + (jnp.exp(-2.0 / (1.0 - jnp.square(s_arg))) / x)[1:],
    )

    return jnp.sum(value * h / 2)


@partial(jit, static_argnums=(0, 1, 2))
def k_lam_jax(L: int, lam: float = 2.0, quad_iters: int = 300) -> float:
    r"""JAX version of k_lam. Compute function :math:`k_{\lambda}` used as a wavelet
        generating function.

    Specifically, this function is derived in [1] and is given by

    .. math::

        k_{\lambda} \equiv \frac{ \int_t^1 \frac{\text{d}t^{\prime}}{t^{\prime}}
        s_{\lambda}^2(t^{\prime})}{ \int_{\frac{1}{\lambda}}^1
        \frac{\text{d}t^{\prime}}{t^{\prime}} s_{\lambda}^2(t^{\prime})},

    where the integrand is defined to be

    .. math::

        s_{\lambda} \equiv s \Big ( \frac{2\lambda}{\lambda - 1}(t-\frac{1}{\lambda})
        - 1 \Big ),

    for infinitely differentiable Cauchy-Schwartz function :math:`s(t) \in C^{\infty}`.

    Args:
        L (int): Harmonic band-limit.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales. Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

        quad_iters (int, optional): Total number of iterations for quadrature
            integration. Defaults to 300.

    Returns:
        (np.ndarray): Value of :math:`k_{\lambda}` computed for values between
            :math:`\frac{1}{\lambda}` and 1, parametrised by :math:`\ell` as required to
            compute the axisymmetric filters in :func:`~tiling_axisym`.

    Note:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the
            sphere", A&A, vol. 558, p. A128, 2013.
    """

    J = samples.j_max(L, lam)

    normalisation = part_scaling_fn(1.0 / lam, 1.0, quad_iters, lam)
    k = jnp.zeros((J + 2, L))

    for j in range(J + 2):
        for l in range(L):
            if l < lam ** (j - 1):
                k = k.at[j, l].set(1.0)
            elif l > lam**j:
                k = k.at[j, l].set(0.0)
            else:
                k = k.at[j, l].set(
                    part_scaling_fn(l / lam**j, 1.0, quad_iters, lam) / normalisation
                )

    return k


def tiling_direction(L: int, N: int = 1) -> np.ndarray:
    r"""Generates the harmonic coefficients for the directionality component of the
        tiling functions.

    Formally, this function implements the follow equation

    .. math::

        _{s}\eta_{\el m} = \nu \vu \sqrt{\frac{1}{2^{\gamma}} \big ( \binom{\gamma}{
                (\gamma - m)/2} \big )}

    which was first derived in `[1] <https://arxiv.org/pdf/1211.1680.pdf>`_.

    Args:
        L (int): Harmonic band-limit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

    Returns:
        np.ndarray: Harmonic coefficients of directionality components
            :math:`_{s}\eta_{\el m}`.

    Notes:
        [1] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint
            arXiv:1509.06749 (2015).
    """
    if N % 2:
        nu = 1
    else:
        nu = 1j

    s_elm = np.zeros((L, 2 * L - 1), dtype=np.complex128)

    for el in range(1, L):
        if (N + el) % 2:
            gamma = min(N - 1, el)
        else:
            gamma = min(N - 1, el - 1)

        for m in range(-el, el + 1):
            if abs(m) < N and (N + m) % 2:
                s_elm[el, L - 1 + m] = nu * np.sqrt(
                    (samples.binomial_coefficient(gamma, ((gamma - m) / 2)))
                    / (2**gamma)
                )
            else:
                s_elm[el, L - 1 + m] = 0.0

    return s_elm


def _spin_normalization(el: int, spin: int = 0) -> float:
    r"""Computes the normalization factor for spin-lowered wavelets, which is
        :math:`\sqrt{\frac{(\ell+s)!}{(\ell-s)!}}`.

    Args:
        el (int): Harmonic index :math:`\ell`.

        spin (int): Spin of field over which to perform the transform. Defaults to 0.

    Returns:
        float: Normalization factor for spin-lowered wavelets.
    """
    factor = 1.0

    for s in range(-abs(spin) + 1, abs(spin) + 1):
        factor *= el + s

    if spin > 0:
        return np.sqrt(factor)
    else:
        return np.sqrt(1.0 / factor)


def _spin_normalization_vectorised(el: np.ndarray, spin: int = 0) -> float:
    r"""Vectorised version of :func:`~_spin_normalization`.
    Args:
        el (int): Harmonic index :math:`\ell`.
        spin (int): Spin of field over which to perform the transform. Defaults to 0.
    Returns:
        float: Normalization factor for spin-lowered wavelets.
    """
    factor = np.arange(-abs(spin) + 1, abs(spin) + 1).reshape(1, 2 * abs(spin) + 1)
    factor = el.reshape(len(el), 1).dot(factor)
    return np.sqrt(np.prod(factor, axis=1) ** (np.sign(spin)))


@partial(jit, static_argnums=(0, 1))
def tiling_direction_jax(L: int, N: int = 1) -> np.ndarray:
    r"""JAX version of tiling_direction. Generates the harmonic coefficients for the
        directionality component of the tiling functions.

    Formally, this function implements the follow equation

    .. math::

        _{s}\eta_{\ell m} = \nu \vu \sqrt{\frac{1}{2^{\gamma}} \big ( \binom{\gamma}{
                (\gamma - m)/2} \big )}

    which was first derived in `[1] <https://arxiv.org/pdf/1211.1680.pdf>`_.

    Args:
        L (int): Harmonic band-limit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

    Returns:
        np.ndarray: Harmonic coefficients of directionality components
            :math:`_{s}\eta_{\ell m}`.

    Notes:
        [1] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint
            arXiv:1509.06749 (2015).
    """

    nu = (N % 2 - 1) ** 2 * 1j + (N % 2)

    s_elm = jnp.zeros((L, 2 * L - 1), dtype=np.complex128)

    for el in range(1, L):
        gamma = min(N - 1, el - 1 + (N + el) % 2)

        ms = jnp.arange(-el, el + 1)
        val = nu * jnp.sqrt(
            (samples.binomial_coefficient_jax(gamma, ((gamma - ms) / 2))) / (2**gamma)
        )

        val = jnp.where(
            (ms < N) & (ms > -N) & ((N + ms) % 2 == 1),
            val,
            jnp.zeros(2 * el + 1),
        )
        s_elm = s_elm.at[el, L - 1 - el : L + el].set(val)

    return s_elm


@partial(jit, static_argnums=(1))
def _spin_normalization_jax(el: np.ndarray, spin: int = 0) -> float:
    r"""JAX version of :func:`~_spin_normalization`.
    Args:
        el (int): Harmonic index :math:`\ell`.
        spin (int): Spin of field over which to perform the transform. Defaults to 0.
    Returns:
        float: Normalization factor for spin-lowered wavelets.
    """
    factor = jnp.arange(-abs(spin) + 1, abs(spin) + 1).reshape(1, 2 * abs(spin) + 1)
    factor = el.reshape(len(el), 1).dot(factor)
    return jnp.sqrt(jnp.prod(factor, axis=1) ** (jnp.sign(spin)))



from jax import jit
import jax.numpy as jnp
import numpy as np
from functools import partial
from typing import Tuple, List
import s2fft
from s2wav import samples
from s2wav.transforms import construct


def synthesis(
    f_wav: jnp.ndarray,
    f_scal: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
    use_c_backend: bool = False,
    _ssht_backend: int = 1,
) -> jnp.ndarray:
    r"""Computes the synthesis directional wavelet transform [1,2].
        Specifically, this transform synthesises the signal :math:`_{s}f(\omega) \in \mathbb{S}^2`
        by summing the contributions from wavelet and scaling coefficients in harmonic space,
        see equation 27 from `[2] <https://arxiv.org/pdf/1509.06749.pdf>`_.

    Args:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (jnp.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) of input signal. Defaults to 0.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh",
            "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if
            sampling="healpix".  Defaults to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (Tuple[jnp.ndarray], optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

        use_c_backend (bool, optional): Execution mode in {"jax" = False, "jax_ssht" = True}.
            Defaults to False.

        _ssht_backend (int, optional, experimental): Whether to default to SSHT core
            (set to 0) recursions or pick up ducc0 (set to 1) accelerated experimental
            backend. Use with caution.

    Raises:
        AssertionError: Shape of wavelet/scaling coefficients incorrect.
        ValueError: If healpix sampling is provided to SSHT C backend.

    Returns:
        jnp.ndarray: Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

    Notes:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the sphere", A&A, vol. 558, p. A128, 2013.
        [2] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint arXiv:1509.06749 (2015).
    """
    if precomps == None and not use_c_backend:
        precomps = construct.generate_wigner_precomputes(
            L, N, J_min, lam, sampling, nside, True, reality
        )
    if use_c_backend and sampling.lower() == "healpix":
        raise ValueError("SSHT C backend does not support healpix sampling.")

    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, True)
    flm = jnp.zeros((L, 2 * L - 1), dtype=jnp.complex128)

    f_scal_lm = (
        s2fft.forward(
            f_scal.real if reality else f_scal,
            Ls,
            spin,
            nside,
            sampling,
            "jax_ssht",
            reality,
            _ssht_backend=_ssht_backend,
        )
        if use_c_backend
        else s2fft.forward_jax(f_scal, Ls, spin, nside, sampling, reality)
    )

    # Sum the all wavelet wigner coefficients for each lmn
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        temp = (
            s2fft.wigner.forward(
                f_wav[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                "jax_ssht",
                reality,
                _ssht_backend=_ssht_backend,
            )
            if use_c_backend
            else s2fft.wigner.forward_jax(
                f_wav[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                reality,
                precomps[j - J_min],
                L_lower=L0j,
            )
        )
        flm = flm.at[L0j:Lj, L - Lj : L - 1 + Lj].add(
            jnp.einsum(
                "ln,nlm->lm",
                filters[0][j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                temp[::2, L0j:, :],
                optimize=True,
            )
        )

    # Sum the all scaling harmonic coefficients for each lm
    phi = filters[1][:Ls] * jnp.sqrt(4 * jnp.pi / (2 * jnp.arange(Ls) + 1))
    flm = flm.at[:Ls, L - Ls : L - 1 + Ls].add(
        jnp.einsum("lm,l->lm", f_scal_lm, phi, optimize=True)
    )
    return (
        s2fft.inverse(
            flm,
            L,
            spin,
            nside,
            sampling,
            "jax_ssht",
            reality,
            _ssht_backend=_ssht_backend,
        )
        if use_c_backend
        else s2fft.inverse_jax(flm, L, spin, nside, sampling, reality)
    )


def analysis(
    f: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
    use_c_backend: bool = False,
    _ssht_backend: int = 1,
) -> Tuple[jnp.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (jnp.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) of input signal. Defaults to 0.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (Tuple[jnp.ndarray], optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

        use_c_backend (bool, optional): Execution mode in {"jax" = False, "jax_ssht" = True}.
            Defaults to False.

        _ssht_backend (int, optional, experimental): Whether to default to SSHT core
            (set to 0) recursions or pick up ducc0 (set to 1) accelerated experimental
            backend. Use with caution.

    Returns:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (jnp.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    if precomps == None and not use_c_backend:
        precomps = construct.generate_wigner_precomputes(
            L, N, J_min, lam, sampling, nside, False, reality
        )
    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, True)

    f_wav_lmn = samples.construct_flmn_jax(L, N, J_min, lam, True)
    f_wav = samples.construct_f_jax(L, N, J_min, lam, sampling, nside, True)

    wav_lm = jnp.einsum(
        "jln, l->jln",
        jnp.conj(filters[0]),
        8 * jnp.pi**2 / (2 * jnp.arange(L) + 1),
        optimize=True,
    )

    flm = (
        s2fft.forward(
            f,
            L,
            spin,
            nside,
            sampling,
            "jax_ssht",
            reality,
            _ssht_backend=_ssht_backend,
        )
        if use_c_backend
        else s2fft.forward_jax(f, L, spin, nside, sampling, reality)
    )

    # Project all wigner coefficients for each lmn onto wavelet coefficients
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        f_wav_lmn[j - J_min] = (
            f_wav_lmn[j - J_min]
            .at[::2, L0j:]
            .add(
                jnp.einsum(
                    "lm,ln->nlm",
                    flm[L0j:Lj, L - Lj : L - 1 + Lj],
                    wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                    optimize=True,
                )
            )
        )

        f_wav[j - J_min] = (
            s2fft.wigner.inverse(
                f_wav_lmn[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                "jax_ssht",
                reality,
                _ssht_backend=_ssht_backend,
            )
            if use_c_backend
            else s2fft.wigner.inverse_jax(
                f_wav_lmn[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                reality,
                precomps[j - J_min],
                False,
                L0j,
            )
        )

    # Project all harmonic coefficients for each lm onto scaling coefficients
    phi = filters[1][:Ls] * jnp.sqrt(4 * jnp.pi / (2 * jnp.arange(Ls) + 1))
    temp = jnp.einsum("lm,l->lm", flm[:Ls, L - Ls : L - 1 + Ls], phi, optimize=True)

    # Handle edge case
    if Ls == 1:
        f_scal = temp * jnp.sqrt(1 / (4 * jnp.pi))
    else:
        f_scal = (
            s2fft.inverse(
                temp,
                Ls,
                spin,
                nside,
                sampling,
                "jax_ssht",
                reality,
                _ssht_backend=_ssht_backend,
            )
            if use_c_backend
            else s2fft.inverse_jax(temp, Ls, spin, nside, sampling, reality)
        )
    return f_wav, f_scal


def flm_to_analysis(
    flm: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    J_max: int = None,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
    use_c_backend: bool = False,
    _ssht_backend: int = 1,
) -> Tuple[jnp.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (jnp.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (jnp.ndarray, optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

        use_c_backend (bool, optional): Execution mode in {"jax" = False, "jax_ssht" = True}.
            Defaults to False.

        _ssht_backend (int, optional, experimental): Whether to default to SSHT core
            (set to 0) recursions or pick up ducc0 (set to 1) accelerated experimental
            backend. Use with caution.

    Returns:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.
    """
    if precomps == None and not use_c_backend:
        precomps = construct.generate_wigner_precomputes(
            L, N, J_min, lam, sampling, nside, False, reality
        )

    J = J_max if J_max is not None else samples.j_max(L, lam)

    f_wav_lmn = samples.construct_flmn_jax(L, N, J_min, lam, True)
    f_wav = samples.construct_f_jax(L, N, J_min, lam, sampling, nside, True)

    wav_lm = jnp.einsum(
        "jln, l->jln",
        jnp.conj(filters),
        8 * jnp.pi**2 / (2 * jnp.arange(L) + 1),
        optimize=True,
    )

    # Project all wigner coefficients for each lmn onto wavelet coefficients
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        f_wav_lmn[j - J_min] = (
            f_wav_lmn[j - J_min]
            .at[::2, L0j:]
            .add(
                jnp.einsum(
                    "lm,ln->nlm",
                    flm[L0j:Lj, L - Lj : L - 1 + Lj],
                    wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                    optimize=True,
                )
            )
        )

        f_wav[j - J_min] = (
            s2fft.wigner.inverse(
                f_wav_lmn[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                "jax_ssht",
                reality,
                _ssht_backend=_ssht_backend,
            )
            if use_c_backend
            else s2fft.wigner.inverse_jax(
                f_wav_lmn[j - J_min],
                Lj,
                Nj,
                nside,
                sampling,
                reality,
                precomps[j - J_min],
                False,
                L0j,
            )
        )

    return f_wav



"""Collection of shared fixtures"""
from functools import partial
from typing import Tuple
import numpy as np
import torch
import pytest
import s2fft
from s2fft import base_transforms as base
from s2fft.sampling import so3_samples
from s2wav import samples

DEFAULT_SEED = 8966433580120847635


def pytest_addoption(parser):
    parser.addoption(
        "--seed",
        type=int,
        nargs="*",
        default=[DEFAULT_SEED],
        help=(
            "Seed(s) to use for random number generator fixture rng in tests. If "
            "multiple seeds are passed tests depending on rng will be run for all "
            "seeds specified."
        ),
    )


def pytest_generate_tests(metafunc):
    if "seed" in metafunc.fixturenames:
        metafunc.parametrize("seed", metafunc.config.getoption("seed"))


def generate_f_wav_scal(
    rng: np.random.Generator,
    L: int,
    N: int,
    J_min: int,
    lam: float,
    sampling: str = "mw",
    reality: bool = False,
    using_torch: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    J = samples.j_max(L, lam)
    flmn = samples.construct_flmn(L, N, J_min, lam, True)
    f_wav_s2let = np.zeros(n_wav(L, N, J_min, lam, True), dtype=np.complex128)
    offset = 0

    f_wav = []
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)

        for n in range(-Nj + 1, Nj, 2):
            for el in range(max(abs(n), L0j), Lj):
                for m in range(-el, el + 1):
                    flmn[j - J_min][Nj - 1 + n, el, Lj - 1 + m] = (
                        rng.uniform() + 1j * rng.uniform()
                    )
        temp = base.wigner.inverse(flmn[j - J_min], Lj, Nj, 0, sampling, reality)

        # Pys2let data entries
        entries = temp.flatten("C")
        f_wav_s2let[offset : offset + len(entries)] = entries
        offset += len(entries)

        # S2wav data entries
        if using_torch:
            temp = torch.from_numpy(temp)
        f_wav.append(temp)

    L_s = samples.scal_bandlimit(L, J_min, lam, True)
    flm = np.zeros((L_s, 2 * L_s - 1), dtype=np.complex128)
    for el in range(L_s):
        for m in range(-el, el + 1):
            flm[el, L_s - 1 + m] = rng.uniform() + 1j * rng.uniform()

    f_scal = base.spherical.inverse(flm, L_s, 0, sampling, reality)

    return (
        f_wav,
        torch.from_numpy(f_scal) if using_torch else f_scal,
        f_wav_s2let,
        f_scal.flatten("C"),
    )


def s2wav_to_s2let(
    f_wav: np.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    using_torch: bool = False,
) -> np.ndarray:

    J = samples.j_max(L, lam)
    f_wav_s2let = np.zeros(n_wav(L, N, J_min, lam, True), dtype=np.complex128)
    offset = 0
    for j in range(J_min, J + 1):
        entries = (
            f_wav[j - J_min].numpy().flatten("C")
            if using_torch
            else f_wav[j - J_min].flatten("C")
        )
        f_wav_s2let[offset : offset + len(entries)] = entries
        offset += len(entries)
    return f_wav_s2let


def n_wav(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    multiresolution: bool = False,
    sampling: str = "mw",
) -> int:

    J = samples.j_max(L, lam)
    count = 0
    for j in range(J_min, J + 1):
        Lj = samples.wav_j_bandlimit(L, j, lam, multiresolution)
        count += np.prod(list(so3_samples.f_shape(Lj, N, sampling)))

    return count


@pytest.fixture
def rng(seed):
    # Import numpy locally to avoid `RuntimeWarning: numpy.ndarray size changed`
    # when importing at module level
    import numpy as np

    return np.random.default_rng(seed)


@pytest.fixture
def f_wav_converter():
    return s2wav_to_s2let


@pytest.fixture
def wavelet_generator(rng):
    return partial(generate_f_wav_scal, rng)


@pytest.fixture
def flm_generator(rng):
    # Import s2fft (and indirectly numpy) locally to avoid
    # `RuntimeWarning: numpy.ndarray size changed` when importing at module level
    return partial(s2fft.utils.signal_generator.generate_flm, rng)



import jax

jax.config.update("jax_enable_x64", True)
import pytest
import numpy as np
from s2wav import filters, samples

L_to_test = [8, 16]
N_to_test = [2, 3]
J_min_to_test = [0]
lam_to_test = [2, 3]


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_axisym_admissibility(L: int, J_min: int, lam: int):
    Psi, Phi = filters.filters_axisym(L, J_min, lam)
    J = samples.j_max(L, lam)
    Psi_j_sum = np.zeros_like(Phi)
    for j in range(J_min, J + 1):
        for el in range(L):
            Psi_j_sum[el] += np.abs(Psi[j, el]) ** 2

    for el in range(L):
        temp = Phi[el] ** 2 + Psi_j_sum[el]
        assert temp == pytest.approx(
            1, rel=1e-14
        ), "Admissibility condition not satisfied at l = " + str(el)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_directional_admissibility(L: int, N: int, J_min: int, lam: int):
    spin = 0
    psi, phi = filters.filters_directional(L, N, J_min, lam)
    J = samples.j_max(L, lam)

    ident = np.zeros(L, dtype=np.complex128)

    for el in range(np.abs(spin), L):
        ident[el] += 4.0 * np.pi / (2 * el + 1) * phi[el] * phi[el]

    for j in range(J + 1):
        for el in range(np.abs(spin), L):
            for m in range(-el, el + 1):
                ident[el] += (
                    8.0
                    * np.pi
                    * np.pi
                    / (2 * el + 1)
                    * psi[j, el, L - 1 + m]
                    * np.conj(psi[j, el, L - 1 + m])
                )

    for el in range(max(np.abs(spin), 1), L):
        assert ident[el] == pytest.approx(
            1 + 0j, rel=1e-14
        ), "Admissibility condition not satisfied at l = " + str(el)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
def test_directional_tiling(L: int, N: int):
    s_elm = filters.tiling_direction(L, N)
    for el in range(1, L):
        temp = 0
        for m in range(-el, el + 1):
            temp += s_elm[el, L - 1 + m] * np.conj(s_elm[el, L - 1 + m])

        assert temp == pytest.approx(
            1 + 0j, rel=1e-14
        ), "Directional tiling satisfied l = " + str(el)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_axisym_vectorised(L: int, J_min: int, lam: int):
    f = filters.filters_axisym(L, J_min, lam)
    f_vect = filters.filters_axisym_vectorised(L, J_min, lam)
    for i in range(2):
        np.testing.assert_allclose(f[i], f_vect[i], rtol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_directional_vectorised(L: int, N: int, J_min: int, lam: int):
    f = filters.filters_directional(L, N, J_min, lam)
    f_vect = filters.filters_directional_vectorised(L, N, J_min, lam)

    for i in range(2):
        np.testing.assert_allclose(f[i], f_vect[i], rtol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_directional_torch(L: int, N: int, J_min: int, lam: int):
    f = filters.filters_directional(L, N, J_min, lam)
    f_vect = filters.filters_directional_vectorised(L, N, J_min, lam, using_torch=True)

    for i in range(2):
        np.testing.assert_allclose(f[i], f_vect[i], rtol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_axisym_jax(L: int, J_min: int, lam: int):
    f = filters.filters_axisym(L, J_min, lam)
    f_jax = filters.filters_axisym_jax(L, J_min, lam)

    for i in range(2):
        np.testing.assert_allclose(f[i], f_jax[i], rtol=1e-13, atol=1e-13)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
def test_directional_jax(L: int, N: int, J_min: int, lam: int):
    f = filters.filters_directional(L, N, J_min, lam)
    f_jax = filters.filters_directional_jax(L, N, J_min, lam)

    for i in range(2):
        np.testing.assert_allclose(f[i], f_jax[i], rtol=1e-13, atol=1e-13)


def test_filter_exceptions():
    L = 8
    with pytest.raises(ValueError) as e:
        filters.filters_axisym(L, 10)

    with pytest.raises(ValueError) as e:
        filters.filters_axisym_vectorised(L, 10)

    with pytest.raises(ValueError) as e:
        filters.filters_axisym_jax(L, 10)



import torch
import numpy as np
import jax.numpy as jnp
from typing import List
import s2fft
from s2fft.precompute_transforms.construct import (
    wigner_kernel_jax,
    spin_spherical_kernel_jax,
)
from s2wav import samples


def generate_full_precomputes(
    L: int,
    N: int,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    forward: bool = False,
    reality: bool = False,
    nospherical: bool = False,
    using_torch: bool = False,
) -> List[jnp.ndarray]:
    r"""Generates a list of precompute arrays associated with the underlying Wigner
    transforms.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh",
            "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        forward (bool, optional): _description_. Defaults to False.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        nospherical (bool, optional): Whether to only compute Wigner precomputes.
            Defaults to False.

        using_torch (bool, optional): Desired frontend functionality. Defaults to False.

    Returns:
        List[jnp.ndarray]: Precomputed recursion arrays for underlying Wigner transforms.
    """
    precomps = []
    J = samples.j_max(L, lam)
    for j in range(J_min, J):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        kernel = wigner_kernel_jax(Lj, Nj, reality, sampling, nside, forward)
        precomps.append(
            torch.from_numpy(np.array(kernel))
        ) if using_torch else precomps.append(kernel)
    Ls = samples.scal_bandlimit(L, J_min, lam, True)
    if nospherical:
        return [], [], precomps
    precompute_scaling = spin_spherical_kernel_jax(
        Ls, 0, reality, sampling, nside, forward
    )
    precompute_full = spin_spherical_kernel_jax(
        L, 0, reality, sampling, nside, not forward
    )

    if using_torch:
        precompute_full = torch.from_numpy(np.array(precompute_full))
        precompute_scaling = torch.from_numpy(np.array(precompute_scaling))

    return precompute_full, precompute_scaling, precomps


def generate_wigner_precomputes(
    L: int,
    N: int,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    forward: bool = False,
    reality: bool = False,
    using_torch: bool = False,
) -> List[jnp.ndarray]:
    r"""Generates a list of precompute arrays associated with the underlying Wigner
    transforms.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh",
            "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        forward (bool, optional): _description_. Defaults to False.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        using_torch (bool, optional): Desired frontend functionality. Defaults to False.

    Returns:
        List[jnp.ndarray]: Precomputed recursion arrays for underlying Wigner transforms.
    """
    precomps = []
    J = samples.j_max(L, lam)
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        kernel = s2fft.generate_precomputes_wigner_jax(
            Lj, Nj, sampling, nside, forward, reality, L0j
        )
        precomps.append(
            torch.from_numpy(np.array(kernel))
        ) if using_torch else precomps.append(kernel)
    return precomps



import jax

jax.config.update("jax_enable_x64", True)
import pytest
import numpy as np
import torch
import pys2let as s2let
from s2fft import base_transforms as sht_base
from s2wav.transforms import (
    wavelet,
    wavelet_precompute,
    wavelet_precompute_torch,
    construct,
)
from s2wav import filters, samples

L_to_test = [8]
N_to_test = [2, 3]
J_min_to_test = [2]
lam_to_test = [2, 3]
reality = [False, True]
sampling_to_test = ["mw", "mwss", "dh", "gl"]
recursive_transform = [False, True]
using_torch_frontend = [False, True]
using_c_backend = [False, True]
_ssht_backends = [0, 1]


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("recursive", recursive_transform)
@pytest.mark.parametrize("using_torch", using_torch_frontend)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_synthesis(
    wavelet_generator,
    L: int,
    N: int,
    J_min: int,
    lam: int,
    reality: bool,
    recursive: bool,
    using_torch: bool,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L, lam)

    # Exceptions
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")
    if recursive and using_torch:
        pytest.skip("Recursive transform not yet available for torch frontend.")
    if not recursive and using_c_backend:
        pytest.skip("Precompute transform not supported from C backend libraries.")

    f_wav, f_scal, f_wav_s2let, f_scal_s2let = wavelet_generator(
        L=L, N=N, J_min=J_min, lam=lam, reality=reality, using_torch=using_torch
    )

    f = s2let.synthesis_wav2px(
        f_wav_s2let,
        f_scal_s2let,
        lam,
        L,
        J_min,
        N,
        spin=0,
        upsample=False,
    )

    filter = filters.filters_directional_vectorised(
        L, N, J_min, lam, using_torch=using_torch
    )
    generator = (
        None
        if using_c_backend
        else (
            construct.generate_wigner_precomputes
            if recursive
            else construct.generate_full_precomputes
        )
    )
    synthesis = (
        wavelet.synthesis
        if recursive
        else (
            wavelet_precompute_torch.synthesis
            if using_torch
            else wavelet_precompute.synthesis
        )
    )
    precomps = (
        None
        if using_c_backend
        else generator(
            L, N, J_min, lam, forward=True, reality=reality, using_torch=using_torch
        )
    )
    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    f_check = synthesis(
        f_wav,
        f_scal,
        L,
        N,
        J_min,
        lam,
        reality=reality,
        filters=filter,
        precomps=precomps,
        **args,
    )

    if using_torch:
        f_check = f_check.resolve_conj().numpy()

    f = np.real(f) if reality else f
    np.testing.assert_allclose(f, f_check.flatten("C"), atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("recursive", recursive_transform)
@pytest.mark.parametrize("using_torch", using_torch_frontend)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_analysis(
    flm_generator,
    f_wav_converter,
    L: int,
    N: int,
    J_min: int,
    lam: int,
    reality: bool,
    recursive: bool,
    using_torch: bool,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L, lam)

    # Exceptions
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")
    if recursive and using_torch:
        pytest.skip("Recursive transform not yet available for torch frontend")
    if not recursive and using_c_backend:
        pytest.skip("Precompute transform not supported from C backend libraries.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality)

    f_wav, f_scal = s2let.analysis_px2wav(
        f.flatten("C").astype(np.complex128), lam, L, J_min, N, spin=0, upsample=False
    )
    filter = filters.filters_directional_vectorised(
        L, N, J_min, lam, using_torch=using_torch
    )
    generator = (
        None
        if using_c_backend
        else (
            construct.generate_wigner_precomputes
            if recursive
            else construct.generate_full_precomputes
        )
    )
    analysis = (
        wavelet.analysis
        if recursive
        else (
            wavelet_precompute_torch.analysis
            if using_torch
            else wavelet_precompute.analysis
        )
    )
    precomps = (
        None
        if using_c_backend
        else generator(
            L, N, J_min, lam, forward=False, reality=reality, using_torch=using_torch
        )
    )

    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    f_wav_check, f_scal_check = analysis(
        torch.from_numpy(f) if using_torch else f,
        L,
        N,
        J_min,
        lam,
        reality=reality,
        filters=filter,
        precomps=precomps,
        **args,
    )

    f_wav_check = f_wav_converter(f_wav_check, L, N, J_min, lam, using_torch)

    np.testing.assert_allclose(f_wav, f_wav_check, atol=1e-14)
    np.testing.assert_allclose(
        f_scal,
        f_scal_check.resolve_conj().numpy().flatten("C")
        if using_torch
        else f_scal_check.flatten("C"),
        atol=1e-14,
    )


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("sampling", sampling_to_test)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_round_trip(
    flm_generator,
    L: int,
    N: int,
    J_min: int,
    lam: int,
    reality: bool,
    sampling: str,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L, lam)

    # Exceptions
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality, sampling=sampling)
    filter = filters.filters_directional_vectorised(L, N, J_min, lam)

    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    f_wav, f_scal = wavelet.analysis(
        f,
        L,
        N,
        J_min,
        lam,
        reality=reality,
        sampling=sampling,
        filters=filter,
        **args,
    )
    f_check = wavelet.synthesis(
        f_wav,
        f_scal,
        L,
        N,
        J_min,
        lam,
        sampling=sampling,
        reality=reality,
        filters=filter,
        **args,
    )

    np.testing.assert_allclose(f, f_check, atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("recursive", recursive_transform)
@pytest.mark.parametrize("using_torch", using_torch_frontend)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_flm_to_analysis(
    flm_generator,
    f_wav_converter,
    L: int,
    N: int,
    J_min: int,
    lam: int,
    reality: bool,
    recursive: bool,
    using_torch: bool,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L, lam)

    # Exceptions
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")
    if recursive and using_torch:
        pytest.skip("Recursive transform not yet available for torch frontend")
    if not recursive and using_c_backend:
        pytest.skip("Precompute transform not supported from C backend libraries.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality)

    f_wav, _ = s2let.analysis_px2wav(
        f.flatten("C").astype(np.complex128), lam, L, J_min, N, spin=0, upsample=False
    )
    filter = filters.filters_directional_vectorised(
        L, N, J_min, lam, using_torch=using_torch
    )[0]

    generator = (
        None
        if using_c_backend
        else (
            construct.generate_wigner_precomputes
            if recursive
            else construct.generate_full_precomputes
        )
    )
    analysis = (
        wavelet.flm_to_analysis
        if recursive
        else (
            wavelet_precompute_torch.flm_to_analysis
            if using_torch
            else wavelet_precompute.flm_to_analysis
        )
    )
    precomps = (
        None
        if using_c_backend
        else generator(
            L, N, J_min, lam, forward=False, reality=reality, using_torch=using_torch
        )
    )

    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    f_wav_check = analysis(
        torch.from_numpy(flm) if using_torch else flm,
        L,
        N,
        J_min,
        None,
        lam,
        reality=reality,
        filters=filter,
        precomps=precomps,
        **args,
    )

    f_wav_check = f_wav_converter(f_wav_check, L, N, J_min, lam, using_torch)

    np.testing.assert_allclose(f_wav, f_wav_check, atol=1e-14)



from jax import jit
import jax.numpy as jnp
from functools import partial
from typing import Tuple, List
from s2fft.precompute_transforms import wigner, spherical
from s2wav import samples


@partial(jit, static_argnums=(2, 3, 4, 5, 6, 7, 8, 9))
def synthesis(
    f_wav: jnp.ndarray,
    f_scal: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
) -> jnp.ndarray:
    r"""Computes the synthesis directional wavelet transform [1,2].
    Specifically, this transform synthesises the signal :math:`_{s}f(\omega) \in \mathbb{S}^2`
    by summing the contributions from wavelet and scaling coefficients in harmonic space,
    see equation 27 from `[2] <https://arxiv.org/pdf/1509.06749.pdf>`_.
    Args:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (jnp.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor
            between consecutive wavelet scales. Note that :math:`\lambda = 2` indicates
            dyadic wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) of input signal. Defaults to 0.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh",
            "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if
            sampling="healpix".  Defaults to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (Tuple[jnp.ndarray], optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

    Raises:
        AssertionError: Shape of wavelet/scaling coefficients incorrect.

    Returns:
        jnp.ndarray: Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

    Notes:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the sphere", A&A, vol. 558, p. A128, 2013.
        [2] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint arXiv:1509.06749 (2015).
    """
    if precomps == None:
        raise ValueError("Must provide precomputed kernels for this transform!")

    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, True)
    flm = jnp.zeros((L, 2 * L - 1), dtype=jnp.complex128)
    f_scal_lm = spherical.forward_transform_jax(
        f_scal, precomps[1], Ls, sampling, reality, spin, nside
    )

    # Sum the all wavelet wigner coefficients for each lmn
    # Note that almost the entire compute is concentrated at the highest two scales.
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        shift = 0 if j < J else -1
        temp = wigner.forward_transform_jax(
            f_wav[j - J_min],
            precomps[2][j - J_min + shift],
            Lj,
            Nj,
            sampling,
            reality,
            nside,
        )
        flm = flm.at[L0j:Lj, L - Lj : L - 1 + Lj].add(
            jnp.einsum(
                "ln,nlm->lm",
                filters[0][j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                temp[::2, L0j:, :],
                optimize=True,
            )
        )

    # Sum the all scaling harmonic coefficients for each lm
    phi = filters[1][:Ls] * jnp.sqrt(4 * jnp.pi / (2 * jnp.arange(Ls) + 1))
    flm = flm.at[:Ls, L - Ls : L - 1 + Ls].add(
        jnp.einsum("lm,l->lm", f_scal_lm, phi, optimize=True)
    )
    return spherical.inverse_transform_jax(
        flm, precomps[0], L, sampling, reality, spin, nside
    )


@partial(jit, static_argnums=(1, 2, 3, 4, 5, 6, 7, 8))
def analysis(
    f: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
) -> Tuple[jnp.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (jnp.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) of input signal. Defaults to 0.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (Tuple[jnp.ndarray], optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

    Returns:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (jnp.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    if precomps == None:
        raise ValueError("Must provide precomputed kernels for this transform!")

    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, True)

    f_wav_lmn = samples.construct_flmn_jax(L, N, J_min, lam, True)
    f_wav = samples.construct_f_jax(L, N, J_min, lam, sampling, nside, True)

    wav_lm = jnp.einsum(
        "jln, l->jln",
        jnp.conj(filters[0]),
        8 * jnp.pi**2 / (2 * jnp.arange(L) + 1),
        optimize=True,
    )

    flm = spherical.forward_transform_jax(
        f, precomps[0], L, sampling, reality, spin, nside
    )
    # Project all wigner coefficients for each lmn onto wavelet coefficients
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        f_wav_lmn[j - J_min] = (
            f_wav_lmn[j - J_min]
            .at[::2, L0j:]
            .add(
                jnp.einsum(
                    "lm,ln->nlm",
                    flm[L0j:Lj, L - Lj : L - 1 + Lj],
                    wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                    optimize=True,
                )
            )
        )
        shift = 0 if j < J else -1
        f_wav[j - J_min] = wigner.inverse_transform_jax(
            f_wav_lmn[j - J_min],
            precomps[2][j - J_min + shift],
            Lj,
            Nj,
            sampling,
            reality,
            nside,
        )

    # Project all harmonic coefficients for each lm onto scaling coefficients
    phi = filters[1][:Ls] * jnp.sqrt(4 * jnp.pi / (2 * jnp.arange(Ls) + 1))
    temp = jnp.einsum("lm,l->lm", flm[:Ls, L - Ls : L - 1 + Ls], phi, optimize=True)
    # Handle edge case
    if Ls == 1:
        f_scal = temp * jnp.sqrt(1 / (4 * jnp.pi))
    else:
        f_scal = spherical.inverse_transform_jax(
            temp, precomps[1], Ls, sampling, reality, spin, nside
        )
    return f_wav, f_scal


@partial(jit, static_argnums=(1, 2, 3, 4, 5, 6, 7, 8))
def flm_to_analysis(
    flm: jnp.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    J_max: int = None,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    filters: Tuple[jnp.ndarray] = None,
    precomps: List[List[jnp.ndarray]] = None,
) -> Tuple[jnp.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (jnp.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        N (int, optional): Upper azimuthal band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

        spin (int, optional): Spin (integer) of input signal. Defaults to 0.

        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.

        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.

        filters (Tuple[jnp.ndarray], optional): Precomputed wavelet filters. Defaults to None.

        precomps (List[jnp.ndarray]): Precomputed list of recursion coefficients. At most
            of length :math:`L^2`, which is a minimal memory overhead.

    Returns:
        f_wav (jnp.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (jnp.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    if precomps == None:
        raise ValueError("Must provide precomputed kernels for this transform!")

    J = J_max if J_max is not None else samples.j_max(L, lam)

    f_wav_lmn = samples.construct_flmn_jax(L, N, J_min, lam, True)
    f_wav = samples.construct_f_jax(L, N, J_min, lam, sampling, nside, True)

    wav_lm = jnp.einsum(
        "jln, l->jln",
        jnp.conj(filters),
        8 * jnp.pi**2 / (2 * jnp.arange(L) + 1),
        optimize=True,
    )

    # Project all wigner coefficients for each lmn onto wavelet coefficients
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, True)
        f_wav_lmn[j - J_min] = (
            f_wav_lmn[j - J_min]
            .at[::2, L0j:]
            .add(
                jnp.einsum(
                    "lm,ln->nlm",
                    flm[L0j:Lj, L - Lj : L - 1 + Lj],
                    wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
                    optimize=True,
                )
            )
        )
        shift = 0 if j < J else -1
        f_wav[j - J_min] = wigner.inverse_transform_jax(
            f_wav_lmn[j - J_min],
            precomps[2][j - J_min + shift],
            Lj,
            Nj,
            sampling,
            reality,
            nside,
        )

    return f_wav



import jax

jax.config.update("jax_enable_x64", True)
import pytest
import jax.numpy as jnp
from jax.test_util import check_grads
import s2fft
from s2wav.transforms import wavelet, wavelet_precompute, construct
from s2wav import filters, samples

L_to_test = [8]
N_to_test = [3]
J_min_to_test = [2]
reality = [False, True]
recursive_transform = [False, True]
using_c_backend = [False, True]
_ssht_backends = [0, 1]


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("recursive", recursive_transform)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_jax_synthesis_gradients(
    wavelet_generator,
    L: int,
    N: int,
    J_min: int,
    reality: bool,
    recursive: bool,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L)

    # Exceptions
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")
    if not recursive and using_c_backend:
        pytest.skip("Precompute transform not supported from C backend libraries.")
    if reality and using_c_backend:
        pytest.skip("Hermitian symmetry for C backend gradients currently conflicts.")

    # Generate random signal
    f_wav, f_scal, _, _ = wavelet_generator(
        L=L, N=N, J_min=J_min, lam=2, reality=reality
    )

    # Generate wavelet filters
    filter = filters.filters_directional_vectorised(L, N, J_min, 2)
    generator = (
        None
        if using_c_backend
        else (
            construct.generate_wigner_precomputes
            if recursive
            else construct.generate_full_precomputes
        )
    )
    synthesis = wavelet.synthesis if recursive else wavelet_precompute.synthesis

    precomps = (
        None
        if using_c_backend
        else generator(L, N, J_min, 2, forward=True, reality=reality)
    )

    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    def func(f_wav, f_scal):
        f = synthesis(
            f_wav,
            f_scal,
            L,
            N,
            J_min,
            reality=reality,
            filters=filter,
            precomps=precomps,
            **args,
        )
        return jnp.sum(jnp.abs(f) ** 2)

    check_grads(func, (f_wav, f_scal), order=1, modes=("rev"))


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("recursive", recursive_transform)
@pytest.mark.parametrize("using_c_backend", using_c_backend)
@pytest.mark.parametrize("_ssht_backend", _ssht_backends)
def test_jax_analysis_gradients(
    flm_generator,
    wavelet_generator,
    L: int,
    N: int,
    J_min: int,
    reality: bool,
    recursive: bool,
    using_c_backend: bool,
    _ssht_backend: int,
):
    J = samples.j_max(L)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")
    if not recursive and using_c_backend:
        pytest.skip("Precompute transform not supported from C backend libraries.")
    if reality and using_c_backend:
        pytest.skip("Hermitian symmetry for C backend gradients currently conflicts.")

    # Generate random signal
    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = s2fft.inverse_jax(flm, L)

    # Generate target signal
    f_wav_target, f_scal_target, _, _ = wavelet_generator(
        L=L, N=N, J_min=J_min, lam=2, reality=reality
    )

    # Generate wavelet filters
    filter = filters.filters_directional_vectorised(L, N, J_min)
    generator = (
        None
        if using_c_backend
        else (
            construct.generate_wigner_precomputes
            if recursive
            else construct.generate_full_precomputes
        )
    )
    analysis = wavelet.analysis if recursive else wavelet_precompute.analysis
    precomps = (
        None
        if using_c_backend
        else generator(L, N, J_min, forward=False, reality=reality)
    )

    args = (
        {"use_c_backend": using_c_backend, "_ssht_backend": _ssht_backend}
        if using_c_backend
        else {}
    )

    def func(f):
        f_wav, f_scal = analysis(
            f, L, N, J_min, reality=reality, filters=filter, precomps=precomps, **args
        )
        loss = jnp.sum(jnp.abs(f_scal - f_scal_target) ** 2)
        for j in range(J - J_min):
            loss += jnp.sum(jnp.abs(f_wav[j - J_min] - f_wav_target[j - J_min]) ** 2)
        return loss

    check_grads(func, (f.real if reality else f,), order=1, modes=("rev"))



import jax

jax.config.update("jax_enable_x64", True)
import pytest
import numpy as np
import pys2let as s2let
from s2fft import base_transforms as sht_base
from s2wav.transforms import base as wav_base
from s2wav import samples

L_to_test = [8]
N_to_test = [2, 3]
J_min_to_test = [2]
lam_to_test = [2, 3]
reality = [False, True]
sampling_to_test = ["mw", "mwss", "dh", "gl"]


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
def test_synthesis_looped(
    wavelet_generator, L: int, N: int, J_min: int, lam: int, reality: bool
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    f_wav, f_scal, f_wav_s2let, f_scal_s2let = wavelet_generator(
        L=L, N=N, J_min=J_min, lam=lam, reality=reality
    )

    f = s2let.synthesis_wav2px(
        f_wav_s2let, f_scal_s2let, lam, L, J_min, N, spin=0, upsample=False
    )

    f_check = wav_base.synthesis_looped(
        f_wav, f_scal, L, N, J_min, lam, reality=reality
    )
    f = np.real(f) if reality else f
    np.testing.assert_allclose(f, f_check.flatten("C"), atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
def test_synthesis_vectorised(
    wavelet_generator, L: int, N: int, J_min: int, lam: int, reality: bool
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    f_wav, f_scal, f_wav_s2let, f_scal_s2let = wavelet_generator(
        L=L, N=N, J_min=J_min, lam=lam, reality=reality
    )

    f = s2let.synthesis_wav2px(
        f_wav_s2let, f_scal_s2let, lam, L, J_min, N, spin=0, upsample=False
    )

    f_check = wav_base.synthesis(f_wav, f_scal, L, N, J_min, lam, reality=reality)
    f = np.real(f) if reality else f
    np.testing.assert_allclose(f, f_check.flatten("C"), atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
def test_analysis_looped(
    flm_generator,
    f_wav_converter,
    L: int,
    N: int,
    J_min: int,
    lam: int,
    reality: bool,
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality)

    f_wav, f_scal = s2let.analysis_px2wav(
        f.flatten("C").astype(np.complex128), lam, L, J_min, N, spin=0, upsample=False
    )
    f_wav_check, f_scal_check = wav_base.analysis_looped(
        f, L, N, J_min, lam, reality=reality
    )
    f_wav_check = f_wav_converter(f_wav_check, L, N, J_min, lam)
    np.testing.assert_allclose(f_wav, f_wav_check, atol=1e-14)
    np.testing.assert_allclose(f_scal, f_scal_check.flatten("C"), atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
def test_analysis_vectorised(
    flm_generator, f_wav_converter, L: int, N: int, J_min: int, lam: int, reality: bool
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality)

    f_wav, f_scal = s2let.analysis_px2wav(
        f.flatten("C").astype(np.complex128), lam, L, J_min, N, spin=0, upsample=False
    )
    f_wav_check, f_scal_check = wav_base.analysis(f, L, N, J_min, lam, reality=reality)

    f_wav_check = f_wav_converter(f_wav_check, L, N, J_min, lam)
    np.testing.assert_allclose(f_wav, f_wav_check.flatten("C"), atol=1e-14)
    np.testing.assert_allclose(f_scal, f_scal_check.flatten("C"), atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("sampling", sampling_to_test)
def test_looped_round_trip(
    flm_generator, L: int, N: int, J_min: int, lam: int, reality: bool, sampling: str
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    nside = int(L / 2)

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(
        flm, L, reality=reality, sampling=sampling, nside=nside
    )

    f_wav, f_scal = wav_base.analysis_looped(
        f, L, N, J_min, lam, reality=reality, sampling=sampling, nside=nside
    )

    f_check = wav_base.synthesis_looped(
        f_wav, f_scal, L, N, J_min, lam, sampling=sampling, reality=reality, nside=nside
    )

    np.testing.assert_allclose(f, f_check, atol=1e-14)


@pytest.mark.parametrize("L", L_to_test)
@pytest.mark.parametrize("N", N_to_test)
@pytest.mark.parametrize("J_min", J_min_to_test)
@pytest.mark.parametrize("lam", lam_to_test)
@pytest.mark.parametrize("reality", reality)
@pytest.mark.parametrize("sampling", sampling_to_test)
def test_vectorised_round_trip(
    flm_generator, L: int, N: int, J_min: int, lam: int, reality: bool, sampling: str
):
    J = samples.j_max(L, lam)
    if J_min >= J:
        pytest.skip("J_min larger than J which isn't a valid test case.")

    flm = flm_generator(L=L, L_lower=0, spin=0, reality=reality)
    f = sht_base.spherical.inverse(flm, L, reality=reality, sampling=sampling)

    f_wav, f_scal = wav_base.analysis(
        f, L, N, J_min, lam, reality=reality, sampling=sampling
    )

    f_check = wav_base.synthesis(
        f_wav, f_scal, L, N, J_min, lam, reality=reality, sampling=sampling
    )

    np.testing.assert_allclose(f, f_check, atol=1e-14)



from jax import jit
import jax.numpy as jnp
import numpy as np
import torch
import math
from functools import partial
from typing import Tuple
from s2fft.sampling import s2_samples, so3_samples
from scipy.special import loggamma
from jax.scipy.special import gammaln as jax_gammaln


def f_scal(
    L: int,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
) -> Tuple[int, int]:
    r"""Computes the shape of scaling coefficients in pixel-space.

    Args:
        L (int): Harmonic bandlimit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw","mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required
            if sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        Tuple[int, int]: Scaling coefficients shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    if multiresolution:
        L_s = min(math.ceil(lam**J_min), L)
    else:
        L_s = L
    return s2_samples.f_shape(L_s, sampling, nside)


def n_wav_scales(L: int, N: int = 1, J_min: int = 0, lam: float = 2.0) -> int:
    r"""Evalutes the total number of wavelet scales.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from {"mw", "mwss", "dh", "gl", "healpix"}.
            Defaults to "mw".

    Returns:
        int: Total number of wavelet scales :math:`n_{j}`.
    """
    return j_max(L, lam) - J_min + 1


def L0_j(j: int, lam: float = 2.0) -> int:
    r"""Computes the minimum harmonic index supported by the given wavelet scale :math:`j`.

    Args:
        j (int): Wavelet scale to consider.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

    Raises:
        ValueError: Kernel type not supported.

    Returns:
        int: The minimum harmonic multipole :math:`el` which is supported by a given wavelet scale.
    """

    return math.ceil(lam ** (j - 1)) if j != 0 else 0


def LN_j(
    L: int,
    j: int = 0,
    N: int = 1,
    lam: float = 2.0,
    multiresolution: bool = False,
) -> Tuple[int, int]:
    r"""Computes the harmonic bandlimit and directionality for scale :math:`j`.

    Args:
        L (int): Harmonic bandlimit.

        j (int): Wavelet scale to consider.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from {"mw", "mwss", "dh", "gl", "healpix"}.
            Defaults to "mw".

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        int: Total number of wavelet scales :math:`n_{j}`.
    """
    if j == 0:
        j += 1
    Lj = wav_j_bandlimit(L, j, lam, multiresolution)
    L0j = L0_j(j, lam) if multiresolution else 0
    Nj = N
    if multiresolution:
        Nj = min(N, Lj)
        Nj += (Nj + N) % 2
    return Lj, Nj, L0j


def f_wav_j(
    L: int,
    j: int = 0,
    N: int = 1,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
) -> Tuple[int, int, int]:
    r"""Computes the shape of wavelet coefficients :math:`f^j` in pixel-space.

    Args:
        L (int): Harmonic bandlimit.

        j (int): Wavelet scale to consider.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw", "mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required
            if sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        Tuple[int, int, int]: Wavelet coefficients shape :math:`[n_{N}, n_{\theta}, n_{\phi}]`.

    Note:
        Each wavelet filter has compact support between a lower and upper harmonic degree.
        Therefore it is possible to conserve information with only a fraction of the
        data per scale :math:`j`. As each scale :math:`j` has a different shape the overall
        wavelet coefficients are stored as a list of arrays, this being the shape of
        one array within such a list.
    """
    Lj, Nj, _ = LN_j(L, j, N, lam, multiresolution)

    return so3_samples.f_shape(Lj, Nj, sampling, nside)


def construct_f(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
    scattering: bool = False,
) -> np.ndarray:
    """Defines a list of arrays corresponding to f_wav.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw", "mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if
            sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        np.ndarray: Empty array (or list of empty arrays) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        f = np.zeros(
            f_wav_j(L, J - 1, N, lam, sampling, nside, multiresolution),
            dtype=np.complex128,
        )
    else:
        f = []
        for j in range(J_min, J + 1):
            f.append(
                np.zeros(
                    f_wav_j(L, j, N, lam, sampling, nside, multiresolution),
                    dtype=np.complex128,
                )
            )
    return f


@partial(jit, static_argnums=(0, 1, 2, 3, 4, 5, 6, 7))
def construct_f_jax(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
    scattering: bool = False,
) -> jnp.ndarray:
    """Defines a list of arrays corresponding to f_wav.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw", "mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if
            sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        jnp.ndarray: Empty array (or list of empty arrays) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        f = jnp.zeros(
            f_wav_j(L, J - 1, N, lam, sampling, nside, multiresolution),
            dtype=jnp.complex128,
        )
    else:
        f = []
        for j in range(J_min, J + 1):
            f.append(
                jnp.zeros(
                    f_wav_j(L, j, N, lam, sampling, nside, multiresolution),
                    dtype=jnp.complex128,
                )
            )
    return f


def construct_f_torch(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
    scattering: bool = False,
) -> torch.tensor:
    """Defines a list of tensors corresponding to f_wav.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw", "mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required if
            sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        torch.tensor: Empty tensor (or list of empty tensors) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        f = torch.zeros(
            f_wav_j(L, J - 1, N, lam, sampling, nside, multiresolution),
            dtype=torch.complex128,
        )
    else:
        f = []
        for j in range(J_min, J + 1):
            f.append(
                torch.zeros(
                    f_wav_j(L, j, N, lam, sampling, nside, multiresolution),
                    dtype=torch.complex128,
                )
            )
    return f


def construct_flm(
    L: int, J_min: int = 0, lam: float = 2.0, multiresolution: bool = False
) -> Tuple[int, int]:
    r"""Returns the shape of scaling coefficients in harmonic space.

    Args:
        L (int): Harmonic bandlimit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        Tuple[int, int]: Scaling coefficients shape :math:`[L, 2L-1]`.
    """
    L_s = scal_bandlimit(L, J_min, lam, multiresolution)
    return np.zeros((L_s, 2 * L_s - 1), dtype=np.complex128)


@partial(jit, static_argnums=(0, 1, 2, 3))
def construct_flm_jax(
    L: int, J_min: int = 0, lam: float = 2.0, multiresolution: bool = False
) -> Tuple[int, int]:
    r"""Returns the shape of scaling coefficients in harmonic space.

    Args:
        L (int): Harmonic bandlimit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        Tuple[int, int]: Scaling coefficients shape :math:`[L, 2L-1]`.
    """
    L_s = scal_bandlimit(L, J_min, lam, multiresolution)
    return jnp.zeros((L_s, 2 * L_s - 1), dtype=jnp.complex128)


def scal_bandlimit(
    L: int, J_min: int = 0, lam: float = 2.0, multiresolution: bool = False
) -> int:
    r"""Returns the harmominc bandlimit of the scaling coefficients.

    Args:
        L (int): Harmonic bandlimit.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        int: Harmonic bandlimit of scaling coefficients.
    """
    if multiresolution:
        return min(math.ceil(lam**J_min), L)
    else:
        return L


def flmn_wav_j(
    L: int,
    j: int,
    N: int = 1,
    lam: float = 2.0,
    multiresolution: bool = False,
) -> Tuple[int, int, int]:
    r"""Returns the shape of wavelet coefficients :math:`f^j_{\ell m n}` in Wigner space.

    Args:
        L (int): Harmonic bandlimit.

        j (int): Wavelet scale to consider.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        Tuple[int, int, int]: Wavelet coefficients shape :math:`[n_{N}, L, 2L-1]`.

    Note:
        Each wavelet filter has compact support between a lower and upper harmonic degree.
        Therefore it is possible to conserve information with only a fraction of the
        data per scale :math:`j`. As each scale :math:`j` has a different shape the overall
        wavelet coefficients are stored as a list of arrays, this being the shape of
        one array within such a list.
    """
    Lj, Nj, _ = LN_j(L, j, N, lam, multiresolution)
    return 2 * Nj - 1, Lj, 2 * Lj - 1


def wav_j_bandlimit(
    L: int, j: int, lam: float = 2.0, multiresolution: bool = False
) -> int:
    r"""Returns the harmominc bandlimit of the scaling coefficients.

    Args:
        L (int): Harmonic bandlimit.

        j (int): Wavelet scale to consider.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

    Returns:
        int: Harmonic bandlimit of scaling coefficients.
    """
    if j == 0:
        j += 1
    if multiresolution:
        return min(math.ceil(lam ** (j + 1)), L)
    else:
        return L


def construct_flmn(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    multiresolution: bool = False,
    scattering: bool = False,
) -> np.ndarray:
    """Defines a list of arrays corresponding to flmn.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        np.ndarray: Empty array (or list of empty arrays) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        flmn = np.zeros(
            flmn_wav_j(L, J - 1, N, lam, multiresolution), dtype=np.complex128
        )
    else:
        flmn = []
        for j in range(J_min, J + 1):
            flmn.append(
                np.zeros(
                    flmn_wav_j(L, j, N, lam, multiresolution),
                    dtype=np.complex128,
                )
            )
    return flmn


@partial(jit, static_argnums=(0, 1, 2, 3, 4, 5))
def construct_flmn_jax(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    multiresolution: bool = False,
    scattering: bool = False,
) -> jnp.ndarray:
    """Defines a list of arrays corresponding to flmn.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        jnp.ndarray: Empty array (or list of empty arrays) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        flmn = jnp.zeros(
            flmn_wav_j(L, J - 1, N, lam, multiresolution), dtype=jnp.complex128
        )
    else:
        flmn = []
        for j in range(J_min, J + 1):
            flmn.append(
                jnp.zeros(
                    flmn_wav_j(L, j, N, lam, multiresolution),
                    dtype=jnp.complex128,
                )
            )
    return flmn


def construct_flmn_torch(
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    multiresolution: bool = False,
    scattering: bool = False,
) -> torch.tensor:
    """Defines a list of tensors corresponding to flmn.

    Args:
        L (int): Harmonic bandlimit.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.

        scattering (bool, optional): Whether to create minimal arrays for scattering transform to
            optimise for memory. Defaults to False.

    Returns:
        torch.tensor: Empty tensor (or list of empty tensors) in which to write data.
    """
    J = j_max(L, lam)
    if scattering:
        flmn = torch.zeros(
            flmn_wav_j(L, J - 1, N, lam, multiresolution), dtype=torch.complex128
        )
    else:
        flmn = []
        for j in range(J_min, J + 1):
            flmn.append(
                torch.zeros(
                    flmn_wav_j(L, j, N, lam, multiresolution),
                    dtype=torch.complex128,
                )
            )
    return flmn


def j_max(L: int, lam: float = 2.0) -> int:
    r"""Computes needlet maximum level required to ensure exact reconstruction.

    Args:
        L (int): Harmonic band-limit.

        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.

    Returns:
        int: The maximum wavelet scale used.
    """
    return int(np.ceil(np.log(L) / np.log(lam)))


def wavelet_shape_check(
    f_w: np.ndarray,
    f_s: np.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    sampling: str = "mw",
    nside: int = None,
    multiresolution: bool = False,
):
    r"""Checks the shape of wavelet coefficients are correct.

    If multiresolution is true, this check will iterate over each scale ensuring that
    each scale :math:`j` is correctly bandlimited at :math:`\lambda^j`.

    Args:
        f_w (np.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_s (np.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.

        L (int): Harmonic bandlimit.

        j (int): Wavelet scale to consider.

        N (int, optional): Upper orientational band-limit. Defaults to 1.

        lam (float, optional): Wavelet parameter which determines the scale factor between
            consecutive wavelet scales. Note that :math:`\lambda = 2` indicates dyadic
            wavelets. Defaults to 2.

        sampling (str, optional): Spherical sampling scheme from
            {"mw", "mwss", "dh", "gl", "healpix"}. Defaults to "mw".

        nside (int, optional): HEALPix Nside resolution parameter.  Only required
            if sampling="healpix".  Defaults to None.

        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to False.
    """
    assert len(f_w) == n_wav_scales(L, N, J_min, lam)
    assert f_s.shape == f_scal(L, J_min, lam, sampling, nside, multiresolution)

    J = j_max(L, lam)
    for j in range(J_min, J + 1):
        assert f_w[j - J_min].shape == f_wav_j(
            L, j, N, lam, sampling, nside, multiresolution
        )


def binomial_coefficient(n: int, k: int) -> int:
    r"""Computes the binomial coefficient :math:`\binom{n}{k}`.

    Args:
        n (int): Number of elements to choose from.

        k (int): Number of elements to pick.

    Returns:
        (int): Number of possible subsets.
    """
    return np.floor(
        0.5 + np.exp(loggamma(n + 1) - loggamma(k + 1) - loggamma(n - k + 1))
    )


def binomial_coefficient_jax(n: int, k: int) -> int:
    r"""Computes the binomial coefficient :math:`\binom{n}{k}`.

    Args:
        n (int): Number of elements to choose from.

        k (int): Number of elements to pick.

    Returns:
        (int): Number of possible subsets.
    """
    return jnp.floor(
        0.5 + jnp.exp(jax_gammaln(n + 1) - jax_gammaln(k + 1) - jax_gammaln(n - k + 1))
    )



import numpy as np
from typing import Tuple
from s2fft import base_transforms as base
from s2wav import samples, filters


def synthesis_looped(
    f_wav: np.ndarray,
    f_scal: np.ndarray,
    L: int,
    N: int,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    multiresolution: bool = True,
) -> np.ndarray:
    r"""Computes the synthesis directional wavelet transform [1,2].
    Specifically, this transform synthesises the signal :math:`_{s}f(\omega) \in \mathbb{S}^2` by summing the contributions from wavelet and scaling coefficients in harmonic space, see equation 27 from `[2] <https://arxiv.org/pdf/1509.06749.pdf>`_.
    Args:
        f_wav (np.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.
        f_scal (np.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
        L (int): Harmonic bandlimit.
        N (int, optional): Upper azimuthal band-limit. Defaults to 1.
        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.
        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.
        spin (int, optional): Spin (integer) of input signal. Defaults to 0.
        spin0 (int, optional): Spin (integer) of output signal. Defaults to 0.
        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".
        nside (int, optional): HEALPix Nside resolution parameter.  Only required
            if sampling="healpix".  Defaults to None.
        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.
        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to True.
    Raises:
        AssertionError: Shape of wavelet/scaling coefficients incorrect.
    Returns:
        np.ndarray: Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.
    Notes:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the sphere", A&A, vol. 558, p. A128, 2013.
        [2] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint arXiv:1509.06749 (2015).
    """

    samples.wavelet_shape_check(
        f_wav, f_scal, L, N, J_min, lam, sampling, nside, multiresolution
    )
    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, multiresolution)
    flm = np.zeros((L, 2 * L - 1), dtype=np.complex128)
    f_scal_lm = base.spherical.forward(f_scal, Ls, spin, sampling, nside, reality)

    # Generate the directional wavelet kernels
    wav_lm, scal_l = filters.filters_directional(L, N, J_min, lam, spin, spin0)

    # Sum the all wavelet wigner coefficients for each lmn
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, multiresolution)
        temp = base.wigner.forward(
            f_wav[j - J_min], Lj, Nj, L0j, sampling, reality, nside
        )
        for n in range(-Nj + 1, Nj, 2):
            for el in range(max(abs(spin), abs(n), L0j), Lj):
                psi = wav_lm[j, el, L - 1 + n]
                for m in range(-el, el + 1):
                    flm[el, L - 1 + m] += temp[Nj - 1 + n, el, Lj - 1 + m] * psi

    # Sum the all scaling harmonic coefficients for each lm
    for el in range(np.abs(spin), Ls):
        phi = np.sqrt(4 * np.pi / (2 * el + 1)) * scal_l[el]

        flm[el, 0 + L - 1] += f_scal_lm[el, Ls - 1 + 0] * phi
        for m in range(1, el + 1):
            flm[el, L - 1 + m] += f_scal_lm[el, Ls - 1 + m] * phi
            if reality:
                flm[el, -m + L - 1] += (-1) ** m * np.conj(flm[el, m + L - 1])
            else:
                flm[el, -m + L - 1] += f_scal_lm[el, Ls - 1 - m] * phi

    return base.spherical.inverse(flm, L, spin, sampling, nside, reality)


def synthesis(
    f_wav: np.ndarray,
    f_scal: np.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    multiresolution: bool = True,
) -> np.ndarray:
    r"""Computes the synthesis directional wavelet transform [1,2].
    Specifically, this transform synthesises the signal :math:`_{s}f(\omega) \in \mathbb{S}^2` by summing the contributions from wavelet and scaling coefficients in harmonic space, see equation 27 from `[2] <https://arxiv.org/pdf/1509.06749.pdf>`_.
    Args:
        f_wav (np.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.
        f_scal (np.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
        L (int): Harmonic bandlimit.
        N (int, optional): Upper azimuthal band-limit. Defaults to 1.
        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.
        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.
        spin (int, optional): Spin (integer) of input signal. Defaults to 0.
        spin0 (int, optional): Spin (integer) of output signal. Defaults to 0.
        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".
        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.
        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.
        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to True.
    Raises:
        AssertionError: Shape of wavelet/scaling coefficients incorrect.
    Returns:
        np.ndarray: Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.
    Notes:
        [1] B. Leidstedt et. al., "S2LET: A code to perform fast wavelet analysis on the sphere", A&A, vol. 558, p. A128, 2013.
        [2] J. McEwen et. al., "Directional spin wavelets on the sphere", arXiv preprint arXiv:1509.06749 (2015).
    """

    samples.wavelet_shape_check(
        f_wav, f_scal, L, N, J_min, lam, sampling, nside, multiresolution
    )

    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, multiresolution)
    flm = np.zeros((L, 2 * L - 1), dtype=np.complex128)
    f_scal_lm = base.spherical.forward(f_scal, Ls, spin, sampling, nside, reality)

    # Generate the directional wavelet kernels
    wav_lm, scal_l = filters.filters_directional_vectorised(
        L, N, J_min, lam, spin, spin0
    )

    # Sum the all wavelet wigner coefficients for each lmn
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, multiresolution)
        temp = base.wigner.forward(
            f_wav[j - J_min], Lj, Nj, L0j, sampling, reality, nside
        )
        flm[L0j:Lj, L - Lj : L - 1 + Lj] += np.einsum(
            "ln,nlm->lm",
            wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
            temp[::2, L0j:, :],
        )

    # Sum the all scaling harmonic coefficients for each lm
    phi = scal_l[:Ls] * np.sqrt(4 * np.pi / (2 * np.arange(Ls) + 1))
    flm[:Ls, L - Ls : L - 1 + Ls] += np.einsum("lm,l->lm", f_scal_lm, phi)

    return base.spherical.inverse(flm, L, spin, sampling, nside, reality)


def analysis_looped(
    f: np.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    multiresolution: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (np.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.
        L (int): Harmonic bandlimit.
        N (int, optional): Upper azimuthal band-limit. Defaults to 1.
        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.
        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.
        spin (int, optional): Spin (integer) of input signal. Defaults to 0.
        spin0 (int, optional): Spin (integer) of output signal. Defaults to 0.
        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".
        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.
        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.
        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to True.

    Returns:
        f_wav (np.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (np.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, multiresolution)

    f_scal_lm = samples.construct_flm(L, J_min, lam, multiresolution)
    f_wav_lmn = samples.construct_flmn(L, N, J_min, lam, multiresolution)
    wav_lm, scal_l = filters.filters_directional(L, N, J_min, lam, spin, spin0)

    flm = base.spherical.forward(f, L, spin, sampling, nside, reality)

    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, multiresolution)
        for n in range(-Nj + 1, Nj, 2):
            for el in range(max(abs(spin), abs(n), L0j), Lj):
                psi = np.conj(wav_lm[j, el, L - 1 + n])
                psi *= 8 * np.pi**2 / (2 * el + 1)

                f_wav_lmn[j - J_min][Nj - 1 + n, el, Lj - 1] = flm[el, L - 1 + 0] * psi
                for m in range(1, el + 1):
                    f_wav_lmn[j - J_min][Nj - 1 + n, el, Lj - 1 + m] = (
                        flm[el, L - 1 + m] * psi
                    )
                    if reality:
                        f_wav_lmn[j - J_min][Nj - 1 - n, el, Lj - 1 - m] = (-1) ** (
                            m + n
                        ) * np.conj(f_wav_lmn[j - J_min][Nj - 1 + n, el, Lj - 1 + m])
                    else:
                        f_wav_lmn[j - J_min][Nj - 1 + n, el, Lj - 1 - m] = (
                            flm[el, L - 1 - m] * psi
                        )

    for el in range(abs(spin), Ls):
        phi = np.sqrt(4.0 * np.pi / (2 * el + 1)) * scal_l[el]

        f_scal_lm[el, Ls - 1 + 0] = flm[el, L - 1 + 0] * phi
        for m in range(1, el + 1):
            f_scal_lm[el, Ls - 1 + m] = flm[el, L - 1 + m] * phi

            if reality:
                f_scal_lm[el, Ls - 1 - m] = (-1) ** m * np.conj(
                    f_scal_lm[el, Ls - 1 + m]
                )
            else:
                f_scal_lm[el, Ls - 1 - m] = flm[el, L - 1 - m] * phi

    f_wav = samples.construct_f(L, N, J_min, lam, sampling, nside, multiresolution)
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, multiresolution)
        f_wav[j - J_min] = base.wigner.inverse(
            f_wav_lmn[j - J_min], Lj, Nj, L0j, sampling, reality, nside
        )

    f_scal = base.spherical.inverse(f_scal_lm, Ls, spin, sampling, nside, reality, 0)
    return f_wav, f_scal


def analysis(
    f: np.ndarray,
    L: int,
    N: int = 1,
    J_min: int = 0,
    lam: float = 2.0,
    spin: int = 0,
    spin0: int = 0,
    sampling: str = "mw",
    nside: int = None,
    reality: bool = False,
    multiresolution: bool = True,
    scattering: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Wavelet analysis from pixel space to wavelet space for complex signals.

    Args:
        f (np.ndarray): Signal :math:`f` on the sphere with shape :math:`[n_{\theta}, n_{\phi}]`.
        L (int): Harmonic bandlimit.
        N (int, optional): Upper azimuthal band-limit. Defaults to 1.
        J_min (int, optional): Lowest frequency wavelet scale to be used. Defaults to 0.
        lam (float, optional): Wavelet parameter which determines the scale factor between consecutive wavelet scales.
            Note that :math:`\lambda = 2` indicates dyadic wavelets. Defaults to 2.
        spin (int, optional): Spin (integer) of input signal. Defaults to 0.
        spin0 (int, optional): Spin (integer) of output signal. Defaults to 0.
        sampling (str, optional): Spherical sampling scheme from {"mw","mwss", "dh", "healpix"}. Defaults to "mw".
        nside (int, optional): HEALPix Nside resolution parameter.  Only required if sampling="healpix".  Defaults
            to None.
        reality (bool, optional): Whether :math:`f \in \mathbb{R}`, if True exploits
            conjugate symmetry of harmonic coefficients. Defaults to False.
        multiresolution (bool, optional): Whether to store the scales at :math:`j_{\text{max}}`
            resolution or its own resolution. Defaults to True.
        scattering (bool, optional): If using for scattering transform return absolute value
            of scattering coefficients.

    Returns:
        f_wav (np.ndarray): Array of wavelet pixel-space coefficients
            with shape :math:`[n_{J}, 2N-1, n_{\theta}, n_{\phi}]`.

        f_scal (np.ndarray): Array of scaling pixel-space coefficients
            with shape :math:`[n_{\theta}, n_{\phi}]`.
    """
    J = samples.j_max(L, lam)
    Ls = samples.scal_bandlimit(L, J_min, lam, multiresolution)

    f_scal_lm = samples.construct_flm(L, J_min, lam, multiresolution)
    f_wav_lmn = samples.construct_flmn(L, N, J_min, lam, multiresolution)
    f_wav = samples.construct_f(L, N, J_min, lam, sampling, multiresolution)

    # Generate the directional wavelet kernels
    wav_lm, scal_l = filters.filters_directional_vectorised(
        L, N, J_min, lam, spin, spin0
    )
    wav_lm = np.einsum(
        "jln, l->jln", np.conj(wav_lm), 8 * np.pi**2 / (2 * np.arange(L) + 1)
    )
    flm = base.spherical.forward(f, L, spin, sampling, nside, reality)

    # Project all wigner coefficients for each lmn onto wavelet coefficients
    # Note that almost the entire compute is concentrated at the highest J
    for j in range(J_min, J + 1):
        Lj, Nj, L0j = samples.LN_j(L, j, N, lam, multiresolution)
        f_wav_lmn[j - J_min][::2, L0j:] = np.einsum(
            "lm,ln->nlm",
            flm[L0j:Lj, L - Lj : L - 1 + Lj],
            wav_lm[j, L0j:Lj, L - Nj : L - 1 + Nj : 2],
        )
        f_wav[j - J_min] = base.wigner.inverse(
            f_wav_lmn[j - J_min], Lj, Nj, L0j, sampling, reality, nside
        )
        if scattering:
            f_wav[j - J_min] = np.abs(f_wav[j - J_min])

    # Project all harmonic coefficients for each lm onto scaling coefficients
    phi = scal_l[:Ls] * np.sqrt(4 * np.pi / (2 * np.arange(Ls) + 1))
    f_scal_lm = np.einsum("lm,l->lm", flm[:Ls, L - Ls : L - 1 + Ls], phi)

    return f_wav, base.spherical.inverse(f_scal_lm, Ls, spin, sampling, nside, reality)



from pathlib import Path
from setuptools import find_packages, setup

this_directory = Path(__file__).parent


long_description = (this_directory / ".pip_readme.rst").read_text()
requirements = (
    (this_directory / "requirements" / "requirements-core.txt").read_text().split("\n")
)


setup(
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    name="s2wav",
    version="1.0.1",
    url="https://github.com/astro-informatics/s2wav",
    author="Authors & Contributors",
    license="GNU General Public License v3 (GPLv3)",
    python_requires=">=3.8",
    install_requires=requirements,
    description=("Differentiable and accelerated wavelet transforms with JAX"),
    long_description_content_type="text/x-rst",
    long_description=long_description,
    packages=find_packages(),
    include_package_data=True,
    pacakge_data={"s2wav": ["default-logging-config.yaml"]},
)



# ~~ Core ~~
from . import filters
from . import samples
from . import transforms

# ~~ Aliases ~~

# JAX recursive transforms
from .transforms.wavelet import analysis, synthesis, flm_to_analysis

# Base transforms
from .transforms.base import analysis as analysis_base
from .transforms.base import synthesis as synthesis_base

# JAX precompute transforms
from .transforms.wavelet_precompute import analysis as analysis_precomp_jax
from .transforms.wavelet_precompute import synthesis as synthesis_precomp_jax

# PyTorch precompute transforms
from .transforms.wavelet_precompute_torch import analysis as analysis_precomp_torch
from .transforms.wavelet_precompute_torch import synthesis as synthesis_precomp_torch

# Martix precompute functions
from .transforms import construct
